#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pandas_ml_utils',
 'pandas_ml_utils.analysis',
 'pandas_ml_utils.datafetching',
 'pandas_ml_utils.extern',
 'pandas_ml_utils.model',
 'pandas_ml_utils.model.features_and_labels',
 'pandas_ml_utils.model.fitting',
 'pandas_ml_utils.summary',
 'pandas_ml_utils.utils',
 'pandas_ml_utils.wrappers']

package_data = \
{'': ['*']}

install_requires = \
['pandas',
 'cachetools',
 'sortedcontainers',
 'scikit-learn',
 'statsmodels',
 'hyperopt',
 'dill']

extras_require = \
{'development': ['sphinx-autodoc-typehints', 'jupyter-sphinx', 'shpinx'],
 'jupyther': ['vdom', 'mako', 'matplotlib', 'seaborn'],
 'reinforcement': ['gym', 'keras', 'keras-rl'],
 'stock_prices': ['yfinance']}

setup(name='pandas-ml-utils',
      version='0.0.18',
      description='Augment pandas DataFrame with methods for machine learning',
      author='KIC',
      author_email='vorarlberger@gmail.com',
      url='https://github.com/KIC/pandas_ml_utils',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>= 3.7',
     )
